@@ -26,7 +26,7 @@
  * along with this program.  If not, see
  * <http://www.gnu.org/licenses/>.
  * 
- * Contact the authors at <mandos@fukt.bsnet.se>.
+ * Contact the authors at <mandos@recompile.se>.
  */
 
 /* Needed by GPGME, specifically gpgme_data_seek() */
@@ -62,7 +62,8 @@
 #include <inttypes.h>		/* PRIu16, PRIdMAX, intmax_t,
 				   strtoimax() */
 #include <assert.h>		/* assert() */
-#include <errno.h>		/* perror(), errno */
+#include <errno.h>		/* perror(), errno,
+				   program_invocation_short_name */
 #include <time.h>		/* nanosleep(), time(), sleep() */
 #include <net/if.h>		/* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
 				   SIOCSIFFLAGS, if_indextoname(),
@@ -73,7 +74,7 @@
 #include <unistd.h>		/* close(), SEEK_SET, off_t, write(),
 				   getuid(), getgid(), seteuid(),
 				   setgid(), pause() */
-#include <arpa/inet.h>		/* inet_pton(), htons */
+#include <arpa/inet.h>		/* inet_pton(), htons, inet_ntop() */
 #include <iso646.h>		/* not, or, and */
 #include <argp.h>		/* struct argp_option, error_t, struct
 				   argp_state, struct argp,
@@ -126,10 +127,21 @@
 bool debug = false;
 static const char mandos_protocol_version[] = "1";
 const char *argp_program_version = "mandos-client " VERSION;
-const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
+const char *argp_program_bug_address = "<mandos@recompile.se>";
 static const char sys_class_net[] = "/sys/class/net";
 char *connect_to = NULL;
 
+/* Doubly linked list that need to be circularly linked when used */
+typedef struct server{
+  const char *ip;
+  uint16_t port;
+  AvahiIfIndex if_index;
+  int af;
+  struct timespec last_seen;
+  struct server *next;
+  struct server *prev;
+} server;
+
 /* Used for passing in values through the Avahi callback functions */
 typedef struct {
   AvahiSimplePoll *simple_poll;
@@ -139,16 +151,25 @@ typedef struct {
   gnutls_dh_params_t dh_params;
   const char *priority;
   gpgme_ctx_t ctx;
+  server *current_server;
 } mandos_context;
 
 /* global context so signal handler can reach it*/
 mandos_context mc = { .simple_poll = NULL, .server = NULL,
 		      .dh_bits = 1024, .priority = "SECURE256"
-		      ":!CTYPE-X.509:+CTYPE-OPENPGP" };
+		      ":!CTYPE-X.509:+CTYPE-OPENPGP",
+		      .current_server = NULL };
 
 sig_atomic_t quit_now = 0;
 int signal_received = 0;
 
+/* Function to use when printing errors */
+void perror_plus(const char *print_text){
+  fprintf(stderr, "Mandos plugin %s: ",
+	  program_invocation_short_name);
+  perror(print_text);
+}
+
 /*
  * Make additional room in "buffer" for at least BUFFER_SIZE more
  * bytes. "buffer_capacity" is how much is currently allocated,
@@ -166,6 +187,44 @@ size_t incbuffer(char **buffer, size_t buffer_length,
   return buffer_capacity;
 }
 
+/* Add server to set of servers to retry periodically */
+int add_server(const char *ip, uint16_t port,
+		 AvahiIfIndex if_index,
+		 int af){
+  int ret;
+  server *new_server = malloc(sizeof(server));
+  if(new_server == NULL){
+    perror_plus("malloc");
+    return -1;
+  }
+  *new_server = (server){ .ip = strdup(ip),
+			 .port = port,
+			 .if_index = if_index,
+			 .af = af };
+  if(new_server->ip == NULL){
+    perror_plus("strdup");
+    return -1;
+  }
+  /* Special case of first server */
+  if (mc.current_server == NULL){
+    new_server->next = new_server;
+    new_server->prev = new_server;
+    mc.current_server = new_server;
+  /* Place the new server last in the list */
+  } else {
+    new_server->next = mc.current_server;
+    new_server->prev = mc.current_server->prev;
+    new_server->prev->next = new_server;
+    mc.current_server->prev = new_server;
+  }
+  ret = clock_gettime(CLOCK_MONOTONIC, &mc.current_server->last_seen);
+  if(ret == -1){
+    perror_plus("clock_gettime");
+    return -1;
+  }
+  return 0;
+}
+
 /* 
  * Initialize GPGME.
  */
@@ -185,7 +244,7 @@ static bool init_gpgme(const char *seckey,
     
     fd = (int)TEMP_FAILURE_RETRY(open(filename, O_RDONLY));
     if(fd == -1){
-      perror("open");
+      perror_plus("open");
       return false;
     }
     
@@ -205,7 +264,7 @@ static bool init_gpgme(const char *seckey,
     
     ret = (int)TEMP_FAILURE_RETRY(close(fd));
     if(ret == -1){
-      perror("close");
+      perror_plus("close");
     }
     gpgme_data_release(pgp_data);
     return true;
@@ -224,7 +283,7 @@ static bool init_gpgme(const char *seckey,
     return false;
   }
   
-    /* Set GPGME home directory for the OpenPGP engine only */
+  /* Set GPGME home directory for the OpenPGP engine only */
   rc = gpgme_get_engine_info(&engine_info);
   if(rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_get_engine_info: %s: %s\n",
@@ -336,7 +395,7 @@ static ssize_t pgp_packet_decrypt(const char *cryptotext,
   
   /* Seek back to the beginning of the GPGME plaintext data buffer */
   if(gpgme_data_seek(dh_plain, (off_t)0, SEEK_SET) == -1){
-    perror("gpgme_data_seek");
+    perror_plus("gpgme_data_seek");
     plaintext_length = -1;
     goto decrypt_end;
   }
@@ -347,7 +406,7 @@ static ssize_t pgp_packet_decrypt(const char *cryptotext,
 				      (size_t)plaintext_length,
 				      plaintext_capacity);
     if(plaintext_capacity == 0){
-	perror("incbuffer");
+	perror_plus("incbuffer");
 	plaintext_length = -1;
 	goto decrypt_end;
     }
@@ -360,7 +419,7 @@ static ssize_t pgp_packet_decrypt(const char *cryptotext,
       break;
     }
     if(ret < 0){
-      perror("gpgme_data_read");
+      perror_plus("gpgme_data_read");
       plaintext_length = -1;
       goto decrypt_end;
     }
@@ -423,12 +482,9 @@ static int init_gnutls_global(const char *pubkeyfilename,
   }
   
   /* OpenPGP credentials */
-  gnutls_certificate_allocate_credentials(&mc.cred);
+  ret = gnutls_certificate_allocate_credentials(&mc.cred);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "GnuTLS memory error: %s\n", /* Spurious warning
-						    from
-						    -Wunreachable-code
-						 */
+    fprintf(stderr, "GnuTLS memory error: %s\n",
 	    safer_gnutls_strerror(ret));
     gnutls_global_deinit();
     return -1;
@@ -589,7 +645,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
   tcp_sd = socket(pf, SOCK_STREAM, 0);
   if(tcp_sd < 0){
     int e = errno;
-    perror("socket");
+    perror_plus("socket");
     errno = e;
     goto mandos_end;
   }
@@ -609,7 +665,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
   }
   if(ret < 0 ){
     int e = errno;
-    perror("inet_pton");
+    perror_plus("inet_pton");
     errno = e;
     goto mandos_end;
   }
@@ -651,7 +707,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
     if(af == AF_INET6 and if_index != AVAHI_IF_UNSPEC){
       char interface[IF_NAMESIZE];
       if(if_indextoname((unsigned int)if_index, interface) == NULL){
-	perror("if_indextoname");
+	perror_plus("if_indextoname");
       } else {
 	fprintf(stderr, "Connection to: %s%%%s, port %" PRIu16 "\n",
 		ip, interface, port);
@@ -671,7 +727,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
 			sizeof(addrstr));
     }
     if(pcret == NULL){
-      perror("inet_ntop");
+      perror_plus("inet_ntop");
     } else {
       if(strcmp(addrstr, ip) != 0){
 	fprintf(stderr, "Canonical address form: %s\n", addrstr);
@@ -692,7 +748,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
   if(ret < 0){
     if ((errno != ECONNREFUSED and errno != ENETUNREACH) or debug){
       int e = errno;
-      perror("connect");
+      perror_plus("connect");
       errno = e;
     }
     goto mandos_end;
@@ -711,7 +767,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
 				   out_size - written));
     if(ret == -1){
       int e = errno;
-      perror("write");
+      perror_plus("write");
       errno = e;
       goto mandos_end;
     }
@@ -742,6 +798,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
     goto mandos_end;
   }
   
+  /* Spurious warning from -Wint-to-pointer-cast */
   gnutls_transport_set_ptr(session, (gnutls_transport_ptr_t) tcp_sd);
   
   if(quit_now){
@@ -784,7 +841,7 @@ static int start_mandos_communication(const char *ip, uint16_t port,
 				   buffer_capacity);
     if(buffer_capacity == 0){
       int e = errno;
-      perror("incbuffer");
+      perror_plus("incbuffer");
       errno = e;
       goto mandos_end;
     }
@@ -895,14 +952,14 @@ static int start_mandos_communication(const char *ip, uint16_t port,
       if(e == 0){
 	e = errno;
       }
-      perror("close");
+      perror_plus("close");
     }
     gnutls_deinit(session);
+    errno = e;
     if(quit_now){
-      e = EINTR;
+      errno = EINTR;
       retval = -1;
     }
-    errno = e;
   }
   return retval;
 }
@@ -951,6 +1008,9 @@ static void resolve_callback(AvahiSServiceResolver *r,
 					   avahi_proto_to_af(proto));
       if(ret == 0){
 	avahi_simple_poll_quit(mc.simple_poll);
+      } else {
+	ret = add_server(ip, port, interface,
+			 avahi_proto_to_af(proto));
       }
     }
   }
@@ -1010,7 +1070,7 @@ static void browse_callback(AvahiSServiceBrowser *b,
   }
 }
 
-/* stop main loop after sigterm has been called */
+/* Signal handler that stops main loop after SIGTERM */
 static void handle_sigterm(int sig){
   if(quit_now){
     return;
@@ -1018,6 +1078,7 @@ static void handle_sigterm(int sig){
   quit_now = 1;
   signal_received = sig;
   int old_errno = errno;
+  /* set main loop to exit */
   if(mc.simple_poll != NULL){
     avahi_simple_poll_quit(mc.simple_poll);
   }
@@ -1032,27 +1093,29 @@ static void handle_sigterm(int sig){
 int good_interface(const struct dirent *if_entry){
   ssize_t ssret;
   char *flagname = NULL;
+  if(if_entry->d_name[0] == '.'){
+    return 0;
+  }
   int ret = asprintf(&flagname, "%s/%s/flags", sys_class_net,
 		     if_entry->d_name);
   if(ret < 0){
-    perror("asprintf");
-    return 0;
-  }
-  if(if_entry->d_name[0] == '.'){
+    perror_plus("asprintf");
     return 0;
   }
   int flags_fd = (int)TEMP_FAILURE_RETRY(open(flagname, O_RDONLY));
   if(flags_fd == -1){
-    perror("open");
+    perror_plus("open");
+    free(flagname);
     return 0;
   }
+  free(flagname);
   typedef short ifreq_flags;	/* ifreq.ifr_flags in netdevice(7) */
   /* read line from flags_fd */
-  ssize_t to_read = (sizeof(ifreq_flags)*2)+3; /* "0x1003\n" */
+  ssize_t to_read = 2+(sizeof(ifreq_flags)*2)+1; /* "0x1003\n" */
   char *flagstring = malloc((size_t)to_read+1); /* +1 for final \0 */
   flagstring[(size_t)to_read] = '\0';
   if(flagstring == NULL){
-    perror("malloc");
+    perror_plus("malloc");
     close(flags_fd);
     return 0;
   }
@@ -1060,7 +1123,7 @@ int good_interface(const struct dirent *if_entry){
     ssret = (ssize_t)TEMP_FAILURE_RETRY(read(flags_fd, flagstring,
 					     (size_t)to_read));
     if(ssret == -1){
-      perror("read");
+      perror_plus("read");
       free(flagstring);
       close(flags_fd);
       return 0;
@@ -1111,6 +1174,14 @@ int good_interface(const struct dirent *if_entry){
     }
     return 0;
   }
+  /* Reject non-ARP interfaces (including dummy interfaces) */
+  if(flags & IFF_NOARP){
+    if(debug){
+      fprintf(stderr, "Rejecting non-ARP interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
   /* Accept this device */
   if(debug){
     fprintf(stderr, "Interface \"%s\" is acceptable\n",
@@ -1119,6 +1190,88 @@ int good_interface(const struct dirent *if_entry){
   return 1;
 }
 
+int notdotentries(const struct dirent *direntry){
+  /* Skip "." and ".." */
+  if(direntry->d_name[0] == '.'
+     and (direntry->d_name[1] == '\0'
+	  or (direntry->d_name[1] == '.'
+	      and direntry->d_name[2] == '\0'))){
+    return 0;
+  }
+  return 1;
+}
+
+int avahi_loop_with_timeout(AvahiSimplePoll *s, int retry_interval){
+  int ret;
+  struct timespec now;
+  struct timespec waited_time;
+  intmax_t block_time;
+  
+  while(true){
+    if(mc.current_server == NULL){
+      if (debug){
+	fprintf(stderr,
+		"Wait until first server is found. No timeout!\n");
+      }
+      ret = avahi_simple_poll_iterate(s, -1);
+    } else {
+      if (debug){
+	fprintf(stderr, "Check current_server if we should run it,"
+		" or wait\n");
+      }
+      /* the current time */
+      ret = clock_gettime(CLOCK_MONOTONIC, &now);
+      if(ret == -1){
+	perror_plus("clock_gettime");
+	return -1;
+      }
+      /* Calculating in ms how long time between now and server
+	 who we visted longest time ago. Now - last seen.  */
+      waited_time.tv_sec = (now.tv_sec
+			    - mc.current_server->last_seen.tv_sec);
+      waited_time.tv_nsec = (now.tv_nsec
+			     - mc.current_server->last_seen.tv_nsec);
+      /* total time is 10s/10,000ms.
+	 Converting to s from ms by dividing by 1,000,
+	 and ns to ms by dividing by 1,000,000. */
+      block_time = ((retry_interval
+		     - ((intmax_t)waited_time.tv_sec * 1000))
+		    - ((intmax_t)waited_time.tv_nsec / 1000000));
+      
+      if (debug){
+	fprintf(stderr, "Blocking for %" PRIdMAX " ms\n", block_time);
+      }
+      
+      if(block_time <= 0){
+	ret = start_mandos_communication(mc.current_server->ip,
+					 mc.current_server->port,
+					 mc.current_server->if_index,
+					 mc.current_server->af);
+	if(ret == 0){
+	  avahi_simple_poll_quit(mc.simple_poll);
+	  return 0;
+	}
+	ret = clock_gettime(CLOCK_MONOTONIC,
+			    &mc.current_server->last_seen);
+	if(ret == -1){
+	  perror_plus("clock_gettime");
+	  return -1;
+	}
+	mc.current_server = mc.current_server->next;
+	block_time = 0; 	/* Call avahi to find new Mandos
+				   servers, but don't block */
+      }
+      
+      ret = avahi_simple_poll_iterate(s, (int)block_time);
+    }
+    if(ret != 0){
+      if (ret > 0 or errno != EINTR) {
+	return (ret != 1) ? ret : 0;
+      }
+    }
+  }
+}
+
 int main(int argc, char *argv[]){
   AvahiSServiceBrowser *sb = NULL;
   int error;
@@ -1141,6 +1294,8 @@ int main(int argc, char *argv[]){
   bool gnutls_initialized = false;
   bool gpgme_initialized = false;
   float delay = 2.5f;
+  double retry_interval = 10; /* 10s between trying a server and
+				 retrying the same server again */
   
   struct sigaction old_sigterm_action = { .sa_handler = SIG_DFL };
   struct sigaction sigterm_action = { .sa_handler = handle_sigterm };
@@ -1152,14 +1307,14 @@ int main(int argc, char *argv[]){
   errno = 0;
   ret = setgid(gid);
   if(ret == -1){
-    perror("setgid");
+    perror_plus("setgid");
   }
   
   /* Lower user privileges (temporarily) */
   errno = 0;
   ret = seteuid(uid);
   if(ret == -1){
-    perror("seteuid");
+    perror_plus("seteuid");
   }
   
   if(quit_now){
@@ -1200,6 +1355,10 @@ int main(int argc, char *argv[]){
 	.arg = "SECONDS",
 	.doc = "Maximum delay to wait for interface startup",
 	.group = 2 },
+      { .name = "retry", .key = 132,
+	.arg = "SECONDS",
+	.doc = "Retry interval used when denied by the mandos server",
+	.group = 2 },
       /*
        * These reproduce what we would get without ARGP_NO_HELP
        */
@@ -1249,6 +1408,14 @@ int main(int argc, char *argv[]){
 	if(errno != 0 or tmp == arg or *tmp != '\0'){
 	  argp_error(state, "Bad delay");
 	}
+      case 132:			/* --retry */
+	errno = 0;
+	retry_interval = strtod(arg, &tmp);
+	if(errno != 0 or tmp == arg or *tmp != '\0'
+	   or (retry_interval * 1000) > INT_MAX
+	   or retry_interval < 0){
+	  argp_error(state, "Bad retry interval");
+	}
 	break;
 	/*
 	 * These reproduce what we would get without ARGP_NO_HELP
@@ -1282,7 +1449,7 @@ int main(int argc, char *argv[]){
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      perror_plus("argp_parse");
       exitcode = EX_OSERR;
       goto end;
     case EINVAL:
@@ -1290,6 +1457,66 @@ int main(int argc, char *argv[]){
       goto end;
     }
   }
+    
+  {
+    /* Work around Debian bug #633582:
+       <http://bugs.debian.org/633582> */
+    struct stat st;
+    
+    /* Re-raise priviliges */
+    errno = 0;
+    ret = seteuid(0);
+    if(ret == -1){
+      perror_plus("seteuid");
+    }
+    
+    if(strcmp(seckey, PATHDIR "/" SECKEY) == 0){
+      int seckey_fd = open(seckey, O_RDONLY);
+      if(seckey_fd == -1){
+	perror_plus("open");
+      } else {
+	ret = (int)TEMP_FAILURE_RETRY(fstat(seckey_fd, &st));
+	if(ret == -1){
+	  perror_plus("fstat");
+	} else {
+	  if(S_ISREG(st.st_mode) and st.st_uid == 0 and st.st_gid == 0){
+	    ret = fchown(seckey_fd, uid, gid);
+	    if(ret == -1){
+	      perror_plus("fchown");
+	    }
+	  }
+	}
+	TEMP_FAILURE_RETRY(close(seckey_fd));
+      }
+    }
+    
+    if(strcmp(pubkey, PATHDIR "/" PUBKEY) == 0){
+      int pubkey_fd = open(pubkey, O_RDONLY);
+      if(pubkey_fd == -1){
+	perror_plus("open");
+      } else {
+	ret = (int)TEMP_FAILURE_RETRY(fstat(pubkey_fd, &st));
+	if(ret == -1){
+	  perror_plus("fstat");
+	} else {
+	  if(S_ISREG(st.st_mode) and st.st_uid == 0 and st.st_gid == 0){
+	    ret = fchown(pubkey_fd, uid, gid);
+	    if(ret == -1){
+	      perror_plus("fchown");
+	    }
+	  }
+	}
+	TEMP_FAILURE_RETRY(close(pubkey_fd));
+      }
+    }
+    
+    /* Lower privileges */
+    errno = 0;
+    ret = seteuid(uid);
+    if(ret == -1){
+      perror_plus("seteuid");
+    }
+  }
   
   if(not debug){
     avahi_set_log_function(empty_log);
@@ -1306,7 +1533,7 @@ int main(int argc, char *argv[]){
 	fprintf(stderr, "Using interface \"%s\"\n", interface);
       }
       if(interface == NULL){
-	perror("malloc");
+	perror_plus("malloc");
 	free(direntries);
 	exitcode = EXIT_FAILURE;
 	goto end;
@@ -1334,19 +1561,19 @@ int main(int argc, char *argv[]){
   sigemptyset(&sigterm_action.sa_mask);
   ret = sigaddset(&sigterm_action.sa_mask, SIGINT);
   if(ret == -1){
-    perror("sigaddset");
+    perror_plus("sigaddset");
     exitcode = EX_OSERR;
     goto end;
   }
   ret = sigaddset(&sigterm_action.sa_mask, SIGHUP);
   if(ret == -1){
-    perror("sigaddset");
+    perror_plus("sigaddset");
     exitcode = EX_OSERR;
     goto end;
   }
   ret = sigaddset(&sigterm_action.sa_mask, SIGTERM);
   if(ret == -1){
-    perror("sigaddset");
+    perror_plus("sigaddset");
     exitcode = EX_OSERR;
     goto end;
   }
@@ -1356,39 +1583,39 @@ int main(int argc, char *argv[]){
   */
   ret = sigaction(SIGINT, NULL, &old_sigterm_action);
   if(ret == -1){
-    perror("sigaction");
+    perror_plus("sigaction");
     return EX_OSERR;
   }
   if(old_sigterm_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGINT, &sigterm_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      perror_plus("sigaction");
       exitcode = EX_OSERR;
       goto end;
     }
   }
   ret = sigaction(SIGHUP, NULL, &old_sigterm_action);
   if(ret == -1){
-    perror("sigaction");
+    perror_plus("sigaction");
     return EX_OSERR;
   }
   if(old_sigterm_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGHUP, &sigterm_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      perror_plus("sigaction");
       exitcode = EX_OSERR;
       goto end;
     }
   }
   ret = sigaction(SIGTERM, NULL, &old_sigterm_action);
   if(ret == -1){
-    perror("sigaction");
+    perror_plus("sigaction");
     return EX_OSERR;
   }
   if(old_sigterm_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGTERM, &sigterm_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      perror_plus("sigaction");
       exitcode = EX_OSERR;
       goto end;
     }
@@ -1411,7 +1638,7 @@ int main(int argc, char *argv[]){
     errno = 0;
     ret = seteuid(0);
     if(ret == -1){
-      perror("seteuid");
+      perror_plus("seteuid");
     }
     
 #ifdef __linux__
@@ -1421,19 +1648,19 @@ int main(int argc, char *argv[]){
     bool restore_loglevel = true;
     if(ret == -1){
       restore_loglevel = false;
-      perror("klogctl");
+      perror_plus("klogctl");
     }
 #endif	/* __linux__ */
     
     sd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
     if(sd < 0){
-      perror("socket");
+      perror_plus("socket");
       exitcode = EX_OSERR;
 #ifdef __linux__
       if(restore_loglevel){
 	ret = klogctl(7, NULL, 0);
 	if(ret == -1){
-	  perror("klogctl");
+	  perror_plus("klogctl");
 	}
       }
 #endif	/* __linux__ */
@@ -1441,19 +1668,19 @@ int main(int argc, char *argv[]){
       errno = 0;
       ret = seteuid(uid);
       if(ret == -1){
-	perror("seteuid");
+	perror_plus("seteuid");
       }
       goto end;
     }
     strcpy(network.ifr_name, interface);
     ret = ioctl(sd, SIOCGIFFLAGS, &network);
     if(ret == -1){
-      perror("ioctl SIOCGIFFLAGS");
+      perror_plus("ioctl SIOCGIFFLAGS");
 #ifdef __linux__
       if(restore_loglevel){
 	ret = klogctl(7, NULL, 0);
 	if(ret == -1){
-	  perror("klogctl");
+	  perror_plus("klogctl");
 	}
       }
 #endif	/* __linux__ */
@@ -1462,7 +1689,7 @@ int main(int argc, char *argv[]){
       errno = 0;
       ret = seteuid(uid);
       if(ret == -1){
-	perror("seteuid");
+	perror_plus("seteuid");
       }
       goto end;
     }
@@ -1472,13 +1699,13 @@ int main(int argc, char *argv[]){
       ret = ioctl(sd, SIOCSIFFLAGS, &network);
       if(ret == -1){
 	take_down_interface = false;
-	perror("ioctl SIOCSIFFLAGS +IFF_UP");
+	perror_plus("ioctl SIOCSIFFLAGS +IFF_UP");
 	exitcode = EX_OSERR;
 #ifdef __linux__
 	if(restore_loglevel){
 	  ret = klogctl(7, NULL, 0);
 	  if(ret == -1){
-	    perror("klogctl");
+	    perror_plus("klogctl");
 	  }
 	}
 #endif	/* __linux__ */
@@ -1486,30 +1713,31 @@ int main(int argc, char *argv[]){
 	errno = 0;
 	ret = seteuid(uid);
 	if(ret == -1){
-	  perror("seteuid");
+	  perror_plus("seteuid");
 	}
 	goto end;
       }
     }
-    /* sleep checking until interface is running */
+    /* Sleep checking until interface is running.
+       Check every 0.25s, up to total time of delay */
     for(int i=0; i < delay * 4; i++){
       ret = ioctl(sd, SIOCGIFFLAGS, &network);
       if(ret == -1){
-	perror("ioctl SIOCGIFFLAGS");
+	perror_plus("ioctl SIOCGIFFLAGS");
       } else if(network.ifr_flags & IFF_RUNNING){
 	break;
       }
       struct timespec sleeptime = { .tv_nsec = 250000000 };
       ret = nanosleep(&sleeptime, NULL);
       if(ret == -1 and errno != EINTR){
-	perror("nanosleep");
+	perror_plus("nanosleep");
       }
     }
     if(not take_down_interface){
       /* We won't need the socket anymore */
       ret = (int)TEMP_FAILURE_RETRY(close(sd));
       if(ret == -1){
-	perror("close");
+	perror_plus("close");
       }
     }
 #ifdef __linux__
@@ -1517,7 +1745,7 @@ int main(int argc, char *argv[]){
       /* Restores kernel loglevel to default */
       ret = klogctl(7, NULL, 0);
       if(ret == -1){
-	perror("klogctl");
+	perror_plus("klogctl");
       }
     }
 #endif	/* __linux__ */
@@ -1527,13 +1755,13 @@ int main(int argc, char *argv[]){
       /* Lower privileges */
       ret = seteuid(uid);
       if(ret == -1){
-	perror("seteuid");
+	perror_plus("seteuid");
       }
     } else {
       /* Lower privileges permanently */
       ret = setuid(uid);
       if(ret == -1){
-	perror("setuid");
+	perror_plus("setuid");
       }
     }
   }
@@ -1555,12 +1783,11 @@ int main(int argc, char *argv[]){
     goto end;
   }
   
-  tempdir_created = true;
   if(mkdtemp(tempdir) == NULL){
-    tempdir_created = false;
-    perror("mkdtemp");
+    perror_plus("mkdtemp");
     goto end;
   }
+  tempdir_created = true;
   
   if(quit_now){
     goto end;
@@ -1608,27 +1835,37 @@ int main(int argc, char *argv[]){
     
     port = (uint16_t)tmpmax;
     *address = '\0';
-    address = connect_to;
     /* Colon in address indicates IPv6 */
     int af;
-    if(strchr(address, ':') != NULL){
+    if(strchr(connect_to, ':') != NULL){
       af = AF_INET6;
+      /* Accept [] around IPv6 address - see RFC 5952 */
+      if(connect_to[0] == '[' and address[-1] == ']')
+	{
+	  connect_to++;
+	  address[-1] = '\0';
+	}
     } else {
       af = AF_INET;
     }
+    address = connect_to;
     
     if(quit_now){
       goto end;
     }
-
+    
     while(not quit_now){
       ret = start_mandos_communication(address, port, if_index, af);
       if(quit_now or ret == 0){
 	break;
       }
-      sleep(15);
-    };
-
+      if(debug){
+	fprintf(stderr, "Retrying in %d seconds\n",
+		(int)retry_interval);
+      }
+      sleep((int)retry_interval);
+    }
+    
     if (not quit_now){
       exitcode = EXIT_SUCCESS;
     }
@@ -1690,8 +1927,13 @@ int main(int argc, char *argv[]){
   if(debug){
     fprintf(stderr, "Starting Avahi loop search\n");
   }
-  
-  avahi_simple_poll_loop(mc.simple_poll);
+
+  ret = avahi_loop_with_timeout(mc.simple_poll,
+				(int)(retry_interval * 1000));
+  if(debug){
+    fprintf(stderr, "avahi_loop_with_timeout exited %s\n",
+	    (ret == 0) ? "successfully" : "with error");
+  }
   
  end:
   
@@ -1718,6 +1960,17 @@ int main(int argc, char *argv[]){
   if(gpgme_initialized){
     gpgme_release(mc.ctx);
   }
+
+  /* Cleans up the circular linked list of Mandos servers the client
+     has seen */
+  if(mc.current_server != NULL){
+    mc.current_server->prev->next = NULL;
+    while(mc.current_server != NULL){
+      server *next = mc.current_server->next;
+      free(mc.current_server);
+      mc.current_server = next;
+    }
+  }
   
   /* Take down the network interface */
   if(take_down_interface){
@@ -1725,59 +1978,46 @@ int main(int argc, char *argv[]){
     errno = 0;
     ret = seteuid(0);
     if(ret == -1){
-      perror("seteuid");
+      perror_plus("seteuid");
     }
     if(geteuid() == 0){
       ret = ioctl(sd, SIOCGIFFLAGS, &network);
       if(ret == -1){
-	perror("ioctl SIOCGIFFLAGS");
+	perror_plus("ioctl SIOCGIFFLAGS");
       } else if(network.ifr_flags & IFF_UP) {
 	network.ifr_flags &= ~(short)IFF_UP; /* clear flag */
 	ret = ioctl(sd, SIOCSIFFLAGS, &network);
 	if(ret == -1){
-	  perror("ioctl SIOCSIFFLAGS -IFF_UP");
+	  perror_plus("ioctl SIOCSIFFLAGS -IFF_UP");
 	}
       }
       ret = (int)TEMP_FAILURE_RETRY(close(sd));
       if(ret == -1){
-	perror("close");
+	perror_plus("close");
       }
       /* Lower privileges permanently */
       errno = 0;
       ret = setuid(uid);
       if(ret == -1){
-	perror("setuid");
+	perror_plus("setuid");
       }
     }
   }
   
-  /* Removes the temp directory used by GPGME */
+  /* Removes the GPGME temp directory and all files inside */
   if(tempdir_created){
-    DIR *d;
-    struct dirent *direntry;
-    d = opendir(tempdir);
-    if(d == NULL){
-      if(errno != ENOENT){
-	perror("opendir");
-      }
-    } else {
-      while(true){
-	direntry = readdir(d);
-	if(direntry == NULL){
-	  break;
-	}
-	/* Skip "." and ".." */
-	if(direntry->d_name[0] == '.'
-	   and (direntry->d_name[1] == '\0'
-		or (direntry->d_name[1] == '.'
-		    and direntry->d_name[2] == '\0'))){
-	  continue;
-	}
+    struct dirent **direntries = NULL;
+    struct dirent *direntry = NULL;
+    int numentries = scandir(tempdir, &direntries, notdotentries,
+			     alphasort);
+    if (numentries > 0){
+      for(int i = 0; i < numentries; i++){
+	direntry = direntries[i];
 	char *fullname = NULL;
 	ret = asprintf(&fullname, "%s/%s", tempdir,
 		       direntry->d_name);
 	if(ret < 0){
-	  perror("asprintf");
+	  perror_plus("asprintf");
 	  continue;
 	}
 	ret = remove(fullname);
@@ -1787,11 +2027,16 @@ int main(int argc, char *argv[]){
 	}
 	free(fullname);
       }
-      closedir(d);
+    }
+
+    /* need to clean even if 0 because man page doesn't specify */
+    free(direntries);
+    if (numentries == -1){
+      perror_plus("scandir");
     }
     ret = rmdir(tempdir);
     if(ret == -1 and errno != ENOENT){
-      perror("rmdir");
+      perror_plus("rmdir");
     }
   }
   
@@ -1802,13 +2047,13 @@ int main(int argc, char *argv[]){
 					    &old_sigterm_action,
 					    NULL));
     if(ret == -1){
-      perror("sigaction");
+      perror_plus("sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      perror_plus("raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());
