@@ -19,7 +19,7 @@
  * along with this program.  If not, see
  * <http://www.gnu.org/licenses/>.
  * 
- * Contact the authors at <mandos@fukt.bsnet.se>.
+ * Contact the authors at <mandos@recompile.se>.
  */
 
 #define _GNU_SOURCE		/* asprintf(), TEMP_FAILURE_RETRY() */
@@ -35,8 +35,9 @@
 #include <sys/types.h>		/* size_t, ssize_t, pid_t, DIR, struct
 				   dirent */
 #include <stddef.h>		/* NULL */
-#include <string.h>		/* strlen(), memcmp() */
-#include <stdio.h>		/* asprintf()*/
+#include <string.h>		/* strlen(), memcmp(), strerror() */
+#include <stdio.h>		/* asprintf(), vasprintf(), vprintf(),
+				   fprintf() */
 #include <unistd.h>		/* close(), write(), readlink(),
 				   read(), STDOUT_FILENO, sleep(),
 				   fork(), setuid(), geteuid(),
@@ -50,11 +51,35 @@
 #include <sys/stat.h>		/* struct stat, lstat(), S_ISLNK */
 #include <sysexits.h>		/* EX_OSERR, EX_UNAVAILABLE */
 #include <argz.h>		/* argz_count(), argz_extract() */
+#include <stdarg.h>		/* va_list, va_start(), ... */
 
 sig_atomic_t interrupted_by_signal = 0;
 int signal_received;
 const char usplash_name[] = "/sbin/usplash";
 
+/* Function to use when printing errors */
+void error_plus(int status, int errnum, const char *formatstring,
+		...){
+  va_list ap;
+  char *text;
+  int ret;
+  
+  va_start(ap, formatstring);
+  ret = vasprintf(&text, formatstring, ap);
+  if (ret == -1){
+    fprintf(stderr, "Mandos plugin %s: ",
+	    program_invocation_short_name);
+    vfprintf(stderr, formatstring, ap);
+    fprintf(stderr, ": ");
+    fprintf(stderr, "%s\n", strerror(errnum));
+    error(status, errno, "vasprintf while printing error");
+    return;
+  }
+  fprintf(stderr, "Mandos plugin ");
+  error(status, errnum, "%s", text);
+  free(text);
+}
+
 static void termination_handler(int signum){
   if(interrupted_by_signal){
     return;
@@ -155,7 +180,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
   size_t cmdline_len = 0;
   DIR *proc_dir = opendir("/proc");
   if(proc_dir == NULL){
-    error(0, errno, "opendir");
+    error_plus(0, errno, "opendir");
     return -1;
   }
   errno = 0;
@@ -183,7 +208,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
       char *exe_link;
       ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
       if(ret == -1){
-	error(0, errno, "asprintf");
+	error_plus(0, errno, "asprintf");
 	goto fail_find_usplash;
       }
       
@@ -195,7 +220,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	  free(exe_link);
 	  continue;
 	}
-	error(0, errno, "lstat");
+	error_plus(0, errno, "lstat");
 	free(exe_link);
 	goto fail_find_usplash;
       }
@@ -226,13 +251,13 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	ret = asprintf(&cmdline_filename, "/proc/%s/cmdline",
 		       proc_ent->d_name);
 	if(ret == -1){
-	  error(0, errno, "asprintf");
+	  error_plus(0, errno, "asprintf");
 	  goto fail_find_usplash;
 	}
 	cl_fd = open(cmdline_filename, O_RDONLY);
 	free(cmdline_filename);
 	if(cl_fd == -1){
-	  error(0, errno, "open");
+	  error_plus(0, errno, "open");
 	  goto fail_find_usplash;
 	}
       }
@@ -244,7 +269,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	if(cmdline_len + blocksize > cmdline_allocated){
 	  tmp = realloc(cmdline, cmdline_allocated + blocksize);
 	  if(tmp == NULL){
-	    error(0, errno, "realloc");
+	    error_plus(0, errno, "realloc");
 	    close(cl_fd);
 	    goto fail_find_usplash;
 	  }
@@ -255,7 +280,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	sret = read(cl_fd, cmdline + cmdline_len,
 		    cmdline_allocated - cmdline_len);
 	if(sret == -1){
-	  error(0, errno, "read");
+	  error_plus(0, errno, "read");
 	  close(cl_fd);
 	  goto fail_find_usplash;
 	}
@@ -263,14 +288,14 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
       } while(sret != 0);
       ret = close(cl_fd);
       if(ret == -1){
-	error(0, errno, "close");
+	error_plus(0, errno, "close");
 	goto fail_find_usplash;
       }
     }
     /* Close directory */
     ret = closedir(proc_dir);
     if(ret == -1){
-      error(0, errno, "closedir");
+      error_plus(0, errno, "closedir");
       goto fail_find_usplash;
     }
     /* Success */
@@ -325,26 +350,26 @@ int main(__attribute__((unused))int argc,
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      error(0, errno, "sigaddset");
+      error_plus(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      error(0, errno, "sigaddset");
+      error_plus(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      error(0, errno, "sigaddset");
+      error_plus(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	error(0, errno, "sigaction");
+	error_plus(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -353,7 +378,7 @@ int main(__attribute__((unused))int argc,
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  error(0, errno, "sigaction");
+	  error_plus(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -362,7 +387,7 @@ int main(__attribute__((unused))int argc,
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	error(0, errno, "sigaction");
+	error_plus(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -371,7 +396,7 @@ int main(__attribute__((unused))int argc,
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  error(0, errno, "sigaction");
+	  error_plus(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -380,7 +405,7 @@ int main(__attribute__((unused))int argc,
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	error(0, errno, "sigaction");
+	error_plus(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -389,7 +414,7 @@ int main(__attribute__((unused))int argc,
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  error(0, errno, "sigaction");
+	  error_plus(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -401,7 +426,7 @@ int main(__attribute__((unused))int argc,
   /* Write command to FIFO */
   if(not usplash_write(&fifo_fd, "TIMEOUT", "0")){
     if(errno != EINTR){
-      error(0, errno, "usplash_write");
+      error_plus(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -413,7 +438,7 @@ int main(__attribute__((unused))int argc,
   
   if(not usplash_write(&fifo_fd, "INPUTQUIET", prompt)){
     if(errno != EINTR){
-      error(0, errno, "usplash_write");
+      error_plus(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -431,7 +456,7 @@ int main(__attribute__((unused))int argc,
   outfifo_fd = open("/dev/.initramfs/usplash_outfifo", O_RDONLY);
   if(outfifo_fd == -1){
     if(errno != EINTR){
-      error(0, errno, "open");
+      error_plus(0, errno, "open");
       status = EX_OSERR;
     }
     goto failure;
@@ -450,7 +475,7 @@ int main(__attribute__((unused))int argc,
       char *tmp = realloc(buf, buf_allocated + blocksize);
       if(tmp == NULL){
 	if(errno != EINTR){
-	  error(0, errno, "realloc");
+	  error_plus(0, errno, "realloc");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -462,7 +487,7 @@ int main(__attribute__((unused))int argc,
 		buf_allocated - buf_len);
     if(sret == -1){
       if(errno != EINTR){
-	error(0, errno, "read");
+	error_plus(0, errno, "read");
 	status = EX_OSERR;
       }
       TEMP_FAILURE_RETRY(close(outfifo_fd));
@@ -477,7 +502,7 @@ int main(__attribute__((unused))int argc,
   ret = close(outfifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      error(0, errno, "close");
+      error_plus(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -490,7 +515,7 @@ int main(__attribute__((unused))int argc,
   
   if(not usplash_write(&fifo_fd, "TIMEOUT", "15")){
     if(errno != EINTR){
-      error(0, errno, "usplash_write");
+      error_plus(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -503,7 +528,7 @@ int main(__attribute__((unused))int argc,
   ret = close(fifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      error(0, errno, "close");
+      error_plus(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -517,7 +542,7 @@ int main(__attribute__((unused))int argc,
       sret = write(STDOUT_FILENO, buf + written, buf_len - written);
       if(sret == -1){
 	if(errno != EINTR){
-	  error(0, errno, "write");
+	  error_plus(0, errno, "write");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -554,7 +579,7 @@ int main(__attribute__((unused))int argc,
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      error(0, errno, "close");
+      error_plus(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -563,7 +588,7 @@ int main(__attribute__((unused))int argc,
   if(outfifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(outfifo_fd));
     if(ret == -1){
-      error(0, errno, "close");
+      error_plus(0, errno, "close");
     }
   }
   
@@ -571,7 +596,7 @@ int main(__attribute__((unused))int argc,
   char **cmdline_argv = malloc((argz_count(cmdline, cmdline_len) + 1)
 			       * sizeof(char *)); /* Count args */
   if(cmdline_argv == NULL){
-    error(0, errno, "malloc");
+    error_plus(0, errno, "malloc");
     return status;
   }
   argz_extract(cmdline, cmdline_len, cmdline_argv); /* Create argv */
@@ -592,13 +617,13 @@ int main(__attribute__((unused))int argc,
        the real user ID (_mandos) */
     ret = setuid(geteuid());
     if(ret == -1){
-      error(0, errno, "setuid");
+      error_plus(0, errno, "setuid");
     }
     
     setsid();
     ret = chdir("/");
     if(ret == -1){
-      error(0, errno, "chdir");
+      error_plus(0, errno, "chdir");
       _exit(EX_OSERR);
     }
 /*     if(fork() != 0){ */
@@ -606,13 +631,13 @@ int main(__attribute__((unused))int argc,
 /*     } */
     ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
     if(ret == -1){
-      error(0, errno, "dup2");
+      error_plus(0, errno, "dup2");
       _exit(EX_OSERR);
     }
     
     execv(usplash_name, cmdline_argv);
     if(not interrupted_by_signal){
-      error(0, errno, "execv");
+      error_plus(0, errno, "execv");
     }
     free(cmdline);
     free(cmdline_argv);
@@ -623,7 +648,7 @@ int main(__attribute__((unused))int argc,
   sleep(2);
   if(not usplash_write(&fifo_fd, "PULSATE", NULL)){
     if(errno != EINTR){
-      error(0, errno, "usplash_write");
+      error_plus(0, errno, "usplash_write");
     }
   }
   
@@ -631,7 +656,7 @@ int main(__attribute__((unused))int argc,
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      error(0, errno, "close");
+      error_plus(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -642,13 +667,13 @@ int main(__attribute__((unused))int argc,
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      error(0, errno, "sigaction");
+      error_plus(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      error(0, errno, "raise");
+      error_plus(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());
